// comp.cpp
// implememtation of the compiler engine.

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <conio.h>
#include "comp.h"
#include "tilang.h"

//#define debugmode

bool compileClrscrStmt(struct programNode *p, FILE *f);
bool compileMainStmt(struct programNode *p, FILE *f);
bool compileGotoxyStmt(struct programNode *p, FILE *f);
bool compileNputsStmt(struct programNode *p, FILE *f);
bool compileSetTextStmt(struct programNode *p, FILE *f);
bool writeProgramPrologue(FILE *f);
bool writeProgramEpilogue(FILE *f);
bool writeFunctions(FILE *f);
bool writeVars(FILE *f);
bool writeConst(FILE *f);
bool makeConst(struct programNode *p);
void newLine(FILE *f);
char *hexString(int x, int y);
void writeNode(struct programNode *p);

struct nameList *constList;
struct nameList *constListHead;
struct nameList *varList;
struct nameList *varListHead;
int indent;

bool CompileTiLang(struct programNode *program, char *outputFileName)
{
  struct programNode *p;
  struct programNode *prev;
  bool result;
  bool done;
  FILE *outputFile;
  indent = 0;

  printf("Output File - %s\n",outputFileName);
  printf("TI-Compiler Compiler Engine v1.0\n");
  p = program;
  done = false;

  outputFile = fopen(outputFileName,"w");
  if (outputFile == NULL) {
	error("Unable to open output file");
	abort();
  }
  writeProgramPrologue(outputFile);
  writeVars(outputFile);
#ifdef debugmode
  writeNode(p);
#endif
  while (p != NULL) {
	switch(p->statementType) {
		case k_clrscr:
			result = compileClrscrStmt(p, outputFile);
			break;
		case k_main:
			result = compileMainStmt(p, outputFile);
			break;
		case k_gotoxy:
			result = compileGotoxyStmt(p, outputFile);
			break;
		case k_nputs:
			result = compileNputsStmt(p, outputFile);
			break;
		case k_settextstyle:
			result = compileSetTextStmt(p, outputFile);
			break;
		}
		if (!result) {
		  error("Compiler Error");
		  return(false);
		}
		if (done) break;
		prev = p;
		p = p->next;
		free(prev);
#ifdef debugmode
		writeNode(p);
#endif
		if (p == NULL) done = true;
	}
	writeFunctions(outputFile);
	writeConst(outputFile);
	writeProgramEpilogue(outputFile);
	fclose(outputFile);
	return(true);
}

// implementation of compiler stuff

bool compileClrscrStmt(struct programNode *p, FILE *f)
{
	switch (p->statementType) {
		case k_clrscr:
			fprintf(f,"ROM_CALL(CLEARLCD)");
			break;
		default: return(false);
	}
	newLine(f);
	return(true);
}

bool compileMainStmt(struct programNode *p, FILE *f)
{

	if (p == NULL) return(false);
	fprintf(f,"; Main Program\n\n");
	fprintf(f,".org 0\n");
	fprintf(f,".db %s,0\n",p->text);
	indent = !indent;
	newLine(f);
	return(true);
}

bool compileGotoxyStmt(struct programNode *p, FILE *f)
{
	if (p == NULL) return(false);
	fprintf(f,"ld hl,$%s",hexString(p->x,p->y));
	newLine(f);
	fprintf(f,"ld (CURSOR_ROW),hl");
	newLine(f);
	return(true);
}

bool compileNputsStmt(struct programNode *p, FILE *f)
{
	if (p == NULL) return(false);
	makeConst(p);
	fprintf(f,"ld de,LB%d",p->x);
	newLine(f);
	fprintf(f,"ld hl,(PROGRAM_ADDR)");
	newLine(f);
	fprintf(f,"add hl,de");
	newLine(f);
	fprintf(f,"ROM_CALL(D_ZT_STR)");
	newLine(f);
	return(true);
}

bool compileSetTextStmt(struct programNode *p, FILE *f)
{
	if (strcmpi(p->text,"B_ON_W") == 0) {
		fprintf(f,"res 3,(IY+05)");
		newLine(f);
	}
	else if (strcmpi(p->text,"W_ON_B") == 0) {
		fprintf(f,"set 3,(IY+05)");
		newLine(f);
	} else {
		error("Incorrect Paramater in call to SetTextStyle");
		return(false);
	}
	return(true);
}

bool writeProgramPrologue(FILE *f)
{
	fprintf(f,"; Compiled with TI-Compiler\n");
	fprintf(f,"; Copyright Samuel Clement 1995\n\n");
	fprintf(f,"#include \"ti-85.h\"\n\n");
	return(true);
}

bool writeProgramEpilogue(FILE *f)
{
	fprintf(f,"\n.end\n");
	return(true);
}

bool writeFunctions(FILE *f)
{
	fprintf(f,"\n\nLibWaitKey:");
	newLine(f);
	fprintf(f,"call GET_KEY");
	newLine(f);
	fprintf(f,"cp $09");
	newLine(f);
	fprintf(f,"jr nz,LibWaitKey");
	newLine(f);
	fprintf(f,"ret");
	newLine(f);
	return(true);
}


bool writeVars(FILE *f)
{
	fprintf(f,"; Variable section\n\n");
	return(true);
}

bool writeConst(FILE *f)
{
	struct nameList *ptr;
	ptr = constListHead;

	fprintf(f,"\n\n; Constant area\n\n");
	while (ptr != NULL) {
		fprintf(f,"LB%d:\n",ptr->value);
		fprintf(f,"     .db %s,0\n",ptr->name);
		ptr = ptr->next;
	}
	return(true);
}

bool makeConst(struct programNode *p)
{
	struct nameList *ptr;

	ptr = allocateMemory(nameList);
	ptr->name = p->text;
	ptr->value = p->x;
	ptr->next = NULL;

	if (constList == NULL) {
		constList = ptr;
		constListHead = constList;
	}
	else {
		constList->next = ptr;
		constList = constList->next;
	}
	return(true);
}

void newLine(FILE *f)
{
	if (indent)
	  fprintf(f,"\n     ");
	else fprintf(f,"\n");
}

char *hexString(int x, int y)
{
	char *temp= {"    \0"};
	int i = -1;
	sprintf(temp,"%2x%2x",x,y);
	while (temp[++i] != EOS)
	  if (temp[i] == ' ') temp[i] = '0';
	return(temp);
}

void writeNode(struct programNode *p)
{
	if (p != NULL) {
		printf("P->StatementType = %d\n",p->statementType);
		printf("P->Text = %s\n",p->text);
		printf("P->X = %d\nP->Y = %d\n",p->x,p->y);
		printf("P->next %p\n\n",p->next);
	} else
		printf("P is NULL!!\n");
}
