// lex.cpp
// implementation of the lexical processor

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <string.h>
#include "tilang.h"
#include "lex.h"

// function declarations

bool readline(FILE *file, char *buffer);
bool getNextToken(char *buffer);
bool identifyToken();
void writeToken(FILE *file);
bool isOneOf(char c, char *list);
bool specStrTok(char *buffer, char *delimiters, char *specTokens,
			bool newBuffer);
bool isTokenKeyword();
bool isTokenIdentifier();
bool isTokenReal();
bool isTokenInteger();
bool isTokenString();
bool isSpecialToken();
bool isTokenDSequence(char *token, int n, int *counter);
bool isTokenCSequence(char *token, int n, int *counter);

// variables
char buffer[255];
FILE *sourceFile;
FILE *tokenFile;
int specIndex;

// delimiters for tokens
char *tokenDelimiters = {" }{;)("};
// special tokens in the language, only one char
char *specialTokens = {"=+-/*,"};
// table of the keywords.. for the language
int numberOfKeywords = 5;
char *keywordTable[] = { "MAIN",
					"CLRSCR",
					"NPUTS",
					"GOTOXY",
					"SETTEXTSTYLE" };

bool LexicalAnalyzer(char *sourceFilename)
{
	// open the source file.
	sourceFile = fopen(sourceFilename,"r");
	if (sourceFile == NULL) {
		error("Cannot open source file");
		return(false);
	}
	// open the token file
	tokenFile = fopen(tokenFileName,"w");
	if (tokenFile == NULL) {
		error("Unable to open output file (token)");
		return(false);
	}
	printf("TI-Complier Token Engine v1.0\n");
	// main tokenizing loop
	while (readline(sourceFile,buffer)) {
		while (getNextToken(buffer)) {

			// get and identify token one at a time
			if (identifyToken())
				writeToken(tokenFile);
			else error("Unknow Symbol");
		}
		tokenType = t_special;
		tokenValue.cval = c_eol;
		writeToken(tokenFile);
    }
    fclose(sourceFile);
    fclose(tokenFile);
    return(true);
}

bool readline(FILE *file, char *buffer)
{
	if (fgets(buffer,255,file) == NULL)
		return(false);
	// get rid of eoline character
	buffer[strlen(buffer)-1] = EOS;
	newLine = true;
	return(true);
}

bool getNextToken(char *buffer)
{
	// first time (ie newline true)
	if (newLine) {
		if (!specStrTok(buffer,tokenDelimiters, specialTokens,true))
			return(false);
		newLine = false;
	}
	else if (!specStrTok(buffer,tokenDelimiters, specialTokens, false))
			return(false);
	return(true);

}

bool specStrTok(char *buffer, char *delimiters,
	char *specTokens, bool newBuffer)
{
	int i;
	// if buffer full, then index = 0
	if (newBuffer) specIndex = 0;

	// determine if first char is special
	if (isOneOf(buffer[specIndex],specialTokens)) {
		tokenBuffer[0] = buffer[specIndex];
		specIndex++;
		return(true);
	}

	// skip delimiters
	while (isOneOf(buffer[specIndex],delimiters) &&
		!(buffer[specIndex] == EOS)) specIndex++;

	// if end of string.. return false
	if (buffer[specIndex] == EOS) return(false);

	// current char special
	if (isOneOf(buffer[specIndex],specTokens)) {
		tokenBuffer[0] = buffer[specIndex];
		specIndex++;
		return(true);
	}

	// if its a string then get it...
	if (buffer[specIndex] == quotationMark)  {
		i = 0;
		tokenBuffer[i++] = buffer[specIndex++];
		while ((buffer[specIndex] != quotationMark) &&
			(buffer[specIndex] != EOS))
			tokenBuffer[i++] = buffer[specIndex++];
		// error checking
		if (buffer[specIndex] == EOS) return(false);
		if (buffer[specIndex] == quotationMark) specIndex++;
		tokenBuffer[i++] = quotationMark;
		tokenBuffer[i] = EOS;
		return(true);
	}

	// if its not a string.. then handle it

	i = 0;
	tokenBuffer[i++] = buffer[specIndex++];
	while (!isOneOf(buffer[specIndex],delimiters) &&
		!isOneOf(buffer[specIndex],specTokens) &&
		(buffer[specIndex] != EOS))
		tokenBuffer[i++] = buffer[specIndex++];
	tokenBuffer[i] = EOS;
	return(true);
}

bool isOneOf(char c, char *list)
{
	unsigned int p;
	for (p=0; p<strlen(list); p++)
		if (c==*(list+p))
			return(true);
	return false;
}

bool identifyToken()
{
	if (isalpha(tokenBuffer[0])) {
		if (!(isTokenKeyword() || isTokenIdentifier()))
			return(false);
		else return(true);
	}
	if (isdigit(tokenBuffer[0])) {
		if (!(isTokenReal() || isTokenInteger()))
			return(false);
		else return(true);
	}
	if (tokenBuffer[0] == quotationMark) {
		if (!isTokenString()) return(false);
		else return(true);
	}
	if (!isSpecialToken()) return(false);
	else return(true);
}

void writeToken(FILE *file)
{
	// write token type for parser
	switch (tokenType) {
		case t_keyword:
		case t_integer:
			fprintf(file,"%d %d\n",tokenType,tokenValue.ival);
			break;
		case t_identifier:
		case t_string:
			fprintf(file,"%d %s\n",tokenType,tokenValue.sval);
			break;
		case t_real:
			fprintf(file,"%d %lf\n",tokenType,tokenValue.dval);
			break;
		case t_special:
			fprintf(file,"%d %c\n",tokenType,tokenValue.cval);
			break;
	}
}

// check for tokens to determine type
// all then functions
bool isTokenKeyword()
{
	int i;
	char cToken[maxTokenLength];
	strcpy(cToken,tokenBuffer);
	strupr(cToken);
	for (i=0; i<numberOfKeywords; i++) {
		if (strcmp(cToken,keywordTable[i]) == 0) {
			tokenValue.ival = i;
			tokenType = t_keyword;
			return(true);
		}
	}
	return(false);
}

bool isTokenIdentifier()
{
	int counter;
	if (!isalpha(tokenBuffer[0])) return(false);
	if (!isTokenCSequence(&tokenBuffer[1],0,&counter)) return(false);
	tokenType = t_identifier;
	strcpy(tokenValue.sval,tokenBuffer);
	return(true);
}

bool isTokenCSequence(char *token, int n, int *counter)
{
	int index;
	*counter = 0;
	index = 0;
	while (ischar(*(token+index))) {
		if (*(token+index) == EOS) return(true);
		index++;
		(*counter)++;
		if (((*counter)>n)&&(n!=0)) return(false);
	}
	return(true);
}

bool isTokenDSequence(char *token, int n, int *counter)
{
	int index;
	*counter = 0;
	index = 0;
	while (isdigit(*(token+index))) {
		index++;
		(*counter)++;
		if (((*counter) > n) && (n!=0)) return(false);
		if (*(token+index)==EOS) return(true);
	}
	return(true);
}

bool isTokenReal()
{
	int counter = 0;
	int index = 0;
	if (isdigit(tokenBuffer[0])) {
		if(!isTokenDSequence(&tokenBuffer[index],0,&counter))
			return(false);
	}
	index += counter;
	if (tokenBuffer[index] == '.') {
		index++;
		if (!isTokenDSequence(&tokenBuffer[index],0,&counter))
			return(false);
	} else return(false);
	index += counter;
	if (tokenBuffer[index] != EOS) return(false);
	tokenType = t_real;
	tokenValue.dval = atof(tokenBuffer);
	return(true);
}

bool isTokenInteger()
{
	int counter=0;
	int index=0;
	if (isdigit(tokenBuffer[0])) {
		if (!isTokenDSequence(&tokenBuffer[0],0,&counter))
			return(false);
	}
	index+=counter;
	// not a real number..
	if (tokenBuffer[index] != EOS) return(false);
	tokenType = t_integer;
	tokenValue.ival = atoi(tokenBuffer);
	return(true);
}

bool isTokenString()
{
	if (tokenBuffer[0] == quotationMark) {
		strcpy(tokenValue.sval,tokenBuffer);
		tokenType = t_string;
		return(true);
	}
	else return(false);
}

bool isSpecialToken()
{
	if (isOneOf(tokenBuffer[0],specialTokens)) {
		tokenValue.cval = tokenBuffer[0];
		tokenType = t_special;
		return(true);
	}
	else	return(false);
}

