// implementation of the parser
// copyright 1995 samuel clement

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>
#include "parse.h"
#include "tilang.h"

//#define debugmode

// token structures
#define maxTokenLength	30

static int tokenType;

struct tokenValueRecord {
	int type;
	union {
		char sval[maxTokenLength];
		int ival;
		double dval;
		char cval;
	};
};

// value of the current token
static struct tokenValueRecord tokenValue;
// buffer to hold the tokens to be processed
static char tokenBuffer[maxTokenLength];

// variables used in parser
char stringBuffer[65];
char variableName[33];
int x, y;
extern FILE *tokenFile;
int lineNumber;

// function declarations for the parser
bool readToken();
struct programNode *bldGotoxyStmt();
struct programNode *bldMainStmt();
struct programNode *bldNputsStmt();
struct programNode *bldClrscrStmt();
struct programNode *bldSetTextStmt();

// implementation of main parser
struct programNode *TiLangParser(struct programNode *program)
{
	struct programNode *p;
	struct programNode *programhead = NULL;
	lineNumber = 0;
	printf("TI-Complier Parsing Engine v2.0\n");
	tokenFile = fopen(tokenFileName,"r");
	if (tokenFile == NULL) {
		error("Cannot open token file for input");
		return(NULL);
	}
	while (readToken()) {
		switch(tokenType) {
			case t_keyword:
				switch(tokenValue.ival) {
					case k_main:
						p = bldMainStmt();
						break;
					case k_gotoxy:
						p = bldGotoxyStmt();
						break;
					case k_clrscr:
						p = bldClrscrStmt();
						break;
					case k_nputs:
						p = bldNputsStmt();
						break;
					case k_settextstyle:
						p = bldSetTextStmt();
						break;
				}
				break;
			default: p = NULL;
		}
		if (p != NULL) {
			if ((program == NULL) && (programhead == NULL)) {
				program = programhead = p;
			} else {
				program->next = p;
				program = program->next;
			}
#ifdef debugmode
			printf("p->statementType = %d\n",p->statementType);
			printf("p->text = %s\n",p->text);
			printf("p->x = %d\np->y = %d\n",p->x, p->y);
			printf("p->next = %s\n\n",p->next);
#endif
		}
	}

	if ((program == NULL) && (programhead == NULL)) {
		error("Unable to parse program");
		return(NULL);
	}
	fclose(tokenFile);
#ifdef debugmode
	p = programhead;
	while (p!=NULL) {
		printf("p->statementType = %d\n",p->statementType);
		printf("p->text = %s\n",p->text);
		printf("p->x = %d\np->y = %d\n",p->x, p->y);
		printf("p->next = %p\n\n",p->next);
		p=p->next;
	}
#endif
	return(programhead);
}

bool readToken()
{
	char buffer[257];
	int i;

	if (fgets(buffer,256,tokenFile) == NULL)
		return(false);
	buffer[strlen(buffer)-1] = EOS;
	lineNumber++;
	sscanf(buffer,"%d",&tokenType);

	switch(tokenType) {
		case t_keyword:
		case t_integer:
			sscanf(buffer,"%d %d",&tokenType,&tokenValue.ival);
			break;
		case t_identifier:
			sscanf(buffer,"%d %s",&tokenType,&tokenValue.sval);
			break;
		case t_real:
			sscanf(buffer,"%d %lf",&tokenType,&tokenValue.dval);
			break;
		case t_string:
			i = 0;
			while (buffer[i++] != BLANK);
			strcpy(tokenValue.sval,&buffer[i]);
			break;
		case t_special:
			sscanf(buffer,"%d %c",&tokenType,&tokenValue.cval);
			break;
	}
	return(true);
}

struct programNode *bldGotoxyStmt()
{
	struct programNode *p;
	p = allocateMemory(programNode);
	p->statementType = k_gotoxy;
	if ((!readToken()) && (tokenType != t_integer)) {
		error("Unable to build (gotoxy) statement");
		return(NULL);
	}
	p->x = tokenValue.ival;
	if ((!readToken()) && (tokenType != t_special)) {
		error("Unable to build (gotoxy) statement");
		return(NULL);
	}
	if ((!readToken()) && (tokenType != t_integer)) {
		error("Unable to build (gotoxy) statement");
		return(NULL);
	}
	p->y = tokenValue.ival;
	p->text = NULL;
	p->next = NULL;
	return(p);
}

struct programNode *bldMainStmt()
{
	struct programNode *p;
	p = allocateMemory(programNode);
	p->statementType = k_main;
	p->x = 0;
	p->y = 0;
	if (!readToken()) {
		error("Unable to build(main) statement");
		return(NULL);
	}
	if (tokenType != t_string) {
		error("Unable to build (main) statement");
		return(NULL);
	}
	strcpy(p->text,tokenValue.sval);
	p->next = NULL;
	return(p);
}

struct programNode *bldNputsStmt()
{
	struct programNode *p;
	p = allocateMemory(programNode);
	p->statementType = k_nputs;
	if ((!readToken()) && (tokenType != t_string)) {
		error("Unable to build (nputs) statement");
		return(NULL);
	}
	p->x = p->y = 0;
	p->x = lineNumber;
	strcpy(p->text,tokenValue.sval);
	p->next = NULL;
	return(p);
}

struct programNode *bldClrscrStmt()
{
	struct programNode *p;
	p = allocateMemory(programNode);
	p->statementType = k_clrscr;
	p->text = NULL;
	p->next = NULL;
	p->x = p->y = 0;
	return(p);
}

struct programNode *bldSetTextStmt()
{
	struct programNode *p;
	p = allocateMemory(programNode);
	p->statementType = k_settextstyle;
	if ((!readToken()) && (tokenType != t_identifier)) {
		error("Unable to build (settextstyle) statement");
		return(NULL);
	}
	strcpy(p->text,tokenValue.sval);
	p->x = p->y = 0;
	p->next = NULL;
	return(p);
}